#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: heartbeat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='heartbeat.proto',
    package='',
    syntax='proto3',
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=
    b'\n\x0fheartbeat.proto\"\x16\n\x07Request\x12\x0b\n\x03tag\x18\x01 \x01(\x0c\"+\n\x05Reply\x12\x0b\n\x03tag\x18\x01 \x01(\x0c\x12\x15\n\rextra_message\x18\x02 \x01(\t2+\n\rGrpcHeartbeat\x12\x1a\n\x04Send\x12\x08.Request\x1a\x06.Reply\"\x00\x62\x06proto3'
)

_REQUEST = _descriptor.Descriptor(
    name='Request',
    full_name='Request',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='tag',
            full_name='Request.tag',
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=19,
    serialized_end=41,
)

_REPLY = _descriptor.Descriptor(
    name='Reply',
    full_name='Reply',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='tag',
            full_name='Reply.tag',
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='extra_message',
            full_name='Reply.extra_message',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=43,
    serialized_end=86,
)

DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Reply'] = _REPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Request = _reflection.GeneratedProtocolMessageType(
    'Request',
    (_message.Message, ),
    {
        'DESCRIPTOR': _REQUEST,
        '__module__': 'heartbeat_pb2'
        # @@protoc_insertion_point(class_scope:Request)
    })
_sym_db.RegisterMessage(Request)

Reply = _reflection.GeneratedProtocolMessageType(
    'Reply',
    (_message.Message, ),
    {
        'DESCRIPTOR': _REPLY,
        '__module__': 'heartbeat_pb2'
        # @@protoc_insertion_point(class_scope:Reply)
    })
_sym_db.RegisterMessage(Reply)

_GRPCHEARTBEAT = _descriptor.ServiceDescriptor(
    name='GrpcHeartbeat',
    full_name='GrpcHeartbeat',
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=88,
    serialized_end=131,
    methods=[
        _descriptor.MethodDescriptor(
            name='Send',
            full_name='GrpcHeartbeat.Send',
            index=0,
            containing_service=None,
            input_type=_REQUEST,
            output_type=_REPLY,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ])
_sym_db.RegisterServiceDescriptor(_GRPCHEARTBEAT)

DESCRIPTOR.services_by_name['GrpcHeartbeat'] = _GRPCHEARTBEAT

# @@protoc_insertion_point(module_scope)
